/* mtrr-uncover: adjust MTRR settings to eliminate nesting in specified range
 *
 * This eliminates a problem running X server drivers on some PCs
 * with 4G or more of RAM.
 *
 * Usage: mtrr-uncover [--help] [--debug] [--mtrrfile mtrrfile] [--execute] [lwb[-upb]]...
 *
 * For each range, the smallest MTRR that covers the specified range will be
 * uncovered (if possible).  If no range is specified, all MTRRs will be
 * uncovered (if possible).
 *
 * There is no guarantee that a solution exists that fits in the
 * number of variable MTRRs in the architecture.
 * Furthermore, the solution may not leave enough MTRRs
 * free for subsequent use.
 *
 * This code uses the /proc/mtrr interface since the ioctls are broken.
 * By default /proc/mtrr is used.  For testing, the --mtrrfile flag allows
 * a different file to be used.
 *
 * The output is a sequence of lines that can be written to /proc/mtrr
 * to fix up the MTRR settings.  They might look something like this:
 *  disable=0
 *  disable=2
 *  base=0x000000000 size=0x080000000 type=write-back
 *  base=0x080000000 size=0x040000000 type=write-back
 *
 * --execute causes these commands to be written to the mtrrfile.
 *
 * See:
 *  kernel/Documentation/x86/mtrr.txt
 *  http://kerneltrap.org/mailarchive/linux-kernel/2007/9/14/260046
 *  https://bugs.launchpad.net/ubuntu/+source/linux-meta/+bug/253204
 *  http://lkml.org/lkml/2008/4/28/52
 *
 * Copyright (C) 2008 D. Hugh Redelmeier
 *
 * $Id: mtrr-uncover.c,v 1.2 2008/09/20 00:41:27 hugh Exp $
 *
 * $Log: mtrr-uncover.c,v $
 * Revision 1.2  2008/09/20 00:41:27  hugh
 * fix target searching
 *
 * Revision 1.1  2008/09/18 20:31:11  hugh
 * Initial revision
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <asm/mtrr.h>
#include <assert.h>

#define FALSE 0
#define TRUE 1

#define elemsof(ar)	(sizeof(ar) / sizeof(*(ar)))	/* elemenst of array */
#define roofof(ar)	(&(ar)[elemsof(ar)])	/* roof of an array */

static const char *me;	/* argv[0] program name */

static int debug = FALSE;

/* Example /proc/mtrr contents:
 *
 * reg00: base=0xd0000000 (3328MB), size= 256MB: uncachable, count=1
 * reg01: base=0xe0000000 (3584MB), size= 512MB: uncachable, count=1
 * reg02: base=0x00000000 (   0MB), size=4096MB: write-back, count=1
 * reg03: base=0x100000000 (4096MB), size=2048MB: write-back, count=1
 * reg04: base=0x180000000 (6144MB), size= 512MB: write-back, count=1
 * reg05: base=0x1a0000000 (6656MB), size= 256MB: write-back, count=1
 */

static const char *pmp = "/proc/mtrr";	/* --mtrrfile <path> */

typedef uint64_t addr_t;	/* representation of address */
#define AF "0x%09" PRIx64	/* address format */

struct target_range {
    struct target_range *previous;
    addr_t lwb;
    addr_t upb;
    struct mtrr *mtrr;	/* MTRR to be uncovered */
};
static struct target_range *targets = NULL;


struct mtrr_type {
    unsigned int type;
    const char *name;
};

const struct mtrr_type mtrr_type[] = {
    { MTRR_TYPE_UNCACHABLE, "uncachable" },	/* must be first for mtrr_type_uncachable */
    { MTRR_TYPE_WRCOMB, "write-combining" },
    { MTRR_TYPE_WRTHROUGH, "write-through" },
    { MTRR_TYPE_WRPROT, "write-protect" },
    { MTRR_TYPE_WRBACK, "write-back" },
    { 0, NULL }
};
#define mtrr_type_uncachable	(&mtrr_type[0])

#define NUM_VAR_MTRR	8

enum flags {
    mtrr_protect = 1 << 0,
    mtrr_target = 1 << 1,
    mtrr_modified = 1 << 2,
    mtrr_deleted = 1 << 3
};

struct mtrr {
    int regno;
    enum flags flags;
    addr_t base;
    addr_t size;
    const struct mtrr_type *type;
    struct mtrr *parent;
    struct mtrr *elder_sib;
    struct mtrr *younger_sib;
    struct mtrr *eldest_child;
};

const static struct mtrr blank_mtrr;	/* all zero and NULL */


/* representations of initial MTRRs
 *
 * Note we stick a notional top MTRR at front.
 */
#define top_mtrr    (&mtrr[0])	/* notional overarching MTRR */
static struct mtrr mtrr[NUM_VAR_MTRR + 1];	/* in order of appearance; none empty */
static int mtrrs_used;	/* number of original MTRRs (including top) */
static struct mtrr *mtrr_index[NUM_VAR_MTRR];	/* index of mtrr[], indexed by register number */
static struct mtrr *sorted_mtrr[NUM_VAR_MTRR + 1];	/* index of mtrr[], sorted */

static inline addr_t
lowbit(addr_t n)
{
    /* trick with two's complement */
    return n & -n;
}

static void
print_mtrr(const struct mtrr *const r, const int depth)
{
    printf("%*s%2d%s%s%s%s " AF "-" AF " %s\n"
	, 3 * depth, "", r->regno
	, (r->flags & mtrr_protect)? "P" : ""
	, (r->flags & mtrr_target)? "T" : ""
	, (r->flags & mtrr_modified)? "'" : ""
	, (r->flags & mtrr_deleted)? "D" : ""
	, r->base, r->base + r->size - 1, r->type->name);
}

static void
print_mtrr_children(const struct mtrr *const r, const int depth)
{
    const struct mtrr*p;

    for (p = r->eldest_child; p != NULL; p = p->younger_sib) {
	print_mtrr(p, depth);
	print_mtrr_children(p, depth+1);
    }
}

static void
slurp(void)
{
    int lino = 0;
    FILE *pm = fopen(pmp, "r");
    char line[256];	/* enough for any line of /proc/mtrr */
    static struct mtrr *mtrr_next = top_mtrr;

    if (pm == NULL) {
	fprintf(stderr, "%s: cannot open %s: %s\n", me, pmp, strerror(errno));
	exit(1);
    }
    /* fill in notional top MTRR */
    {
	struct mtrr *r = mtrr_next++;

	r->regno = -1;
	r->base = (addr_t)0;
	r->size = ~(~(addr_t)0 >> 1);	/* only high bit on */
	r->type = mtrr_type_uncachable;
	r->parent = r->elder_sib = r->younger_sib = r->eldest_child = NULL;
	r->flags |= mtrr_protect;
    }
    while (fgets(line, sizeof(line), pm) != NULL) {
	const char *lp = line;
	const char *pat = "regIr: base=Ib (SBB), size=SsB: T, count=Ic";
	unsigned long long  n;
	unsigned int regno;
	addr_t base;
	addr_t size;
	const struct mtrr_type *type;
	unsigned int count;

	lino++;
	if (*lp == '#' && line[strlen(line) - 1] == '\n') {
	    if (debug)
		printf("%s", line);
	    continue;	/* treat as a comment -- convenient for testfiles */
	}

	while (*pat != '\0') {
	    char p = *pat++;

	    switch (p) {
	    case 'I':	/* Integer */
	    case 'S':	/* Scaled integer */
		{
		    char *ep;

		    /* decode integer */
		    errno = 0;
		    n = strtoull(lp, &ep, 0);
		    if (errno != 0 || ep == lp) {
			fprintf(stderr
			    , "%s: error reading %s line %d: expected a number\n"
			    , me, pmp, lino);
			exit(1);
		    }
		    lp = ep;

		    /* handle scale factor, if specified */
		    if (p == 'S') {
			switch (*lp) {
			case 'M':
			    n <<= 20;
			    break;

			case 'K':
			    n <<= 10;
			    break;

			default:
			    fprintf(stderr
				, "%s: error reading %s line %d: expected M or K but found '%c'\n"
				, me, pmp, lino, *lp);
			    exit(1);
			}
			lp++;
		    }

		    /* consume decoded number */
		    switch (*pat++) {
		    case 'r':	/* store Region number */
			if (n >= NUM_VAR_MTRR) {
			    fprintf(stderr
				, "%s: error reading %s line %d: region number too large\n"
				, me, pmp, lino);
			    exit(1);
			}
			regno = n;
			if (mtrr_index[regno] != NULL) {
			    fprintf(stderr
				, "%s: error reading %s line %d: region number %d is repeated\n"
				, me, pmp, lino, regno);
			    exit(1);
			}
			break;

		    case 'b':	/* store Base, in hex */
			base = n;
			break;

		    case 'B':	/* check redundant Base, in decimal, with scale */
			if (base != n) {
			    fprintf(stderr
				, "%s: error reading %s line %d: bases don't match\n"
				, me, pmp, lino);
			    exit(1);
			}
			break;

		    case 's':	/* store Size */
			size = n;
			break;

		    case 'c':	/* store Count */
			count = n;
			break;

		    default:
			fprintf(stderr
			    , "%s: error reading %s line %d: invalid pattern character '%c'\n"
			    , me, pmp, lino, pat[-1]);
		    }
		}
		break;

	    case 'T':	/* Type */
		{
		    const struct mtrr_type *mt;

		    for (mt = mtrr_type; ; mt++ ) {
			if (mt->name == NULL) {
			    fprintf(stderr
				, "%s: error reading %s line %d: unknown type\n"
				, me, pmp, lino);
			    exit(1);
			}
			if (strncmp(lp, mt->name, strlen(mt->name)) == 0) {
			    type = mt;
			    lp += strlen(mt->name);
			    break;
			}
		    }
		}
		break;

	    default:
		if (*lp != p) {
		    fprintf(stderr
			, "%s: error reading %s line %d: expected '%c' but found '%c'\n"
			, me, pmp, lino, p, *lp);
		    exit(1);
		}
		lp++;
		break;
	    }
	}
	if (*lp != '\n') {
	    fprintf(stderr
		, "%s: error reading %s line %d: expected newline\n"
		, me, pmp, lino);
	    exit(1);
	}
	if (debug)
	    printf("Region %d: base: " AF ", size: " AF ", type: %s, count: %u\n"
	    , regno, base, size, type->name, count);
	if (lowbit(size) != size) {
	    fprintf(stderr
		, "%s: error reading %s line %d: size is not a power of two\n"
		, me, pmp, lino);
	    exit(1);
	} 
	if (size << 1 == 0) {
	    /* Must not be as large as top MTRR's size.
	     * But that wouldn't be legal anyway.
	     */
	    fprintf(stderr
		, "%s: error reading %s line %d: size is too big\n"
		, me, pmp, lino);
	    exit(1);
	}
	if (base % size != 0) {
	    fprintf(stderr
		, "%s: error reading %s line %d: base is not a multiple of size\n"
		, me, pmp, lino);
	    exit(1);
	}
	{
	    struct mtrr *const r = mtrr_next++;

	    if (count != 1) {
		fprintf(stderr, "%s: warning: MTRR %d has count %d so I cannot change it.  Perhaps I should be run earlier.\n"
		    , me, regno, count);
		r->flags |= mtrr_protect;
	    }
	    r->regno = regno;
	    r->base = base;
	    r->size = size;
	    r->type = type;
	    r->parent = r->elder_sib = r->younger_sib = r->eldest_child = NULL;
	    mtrr_index[regno] = r;
	}
    }
    if (ferror(pm)) {
	fprintf(stderr, "%s: error reading %s: %s\n", me, pmp, strerror(errno));
	exit(1);
    }
    fclose(pm);
    mtrrs_used = mtrr_next - mtrr;
}

static inline int
inside(const struct mtrr *r, addr_t a)
{
    return (a & ~(r->size - 1)) == r->base;
}

/* compare function for regions:
 * order by base; when base is equal, larger size comes first
 */
static int
reg_comp(const void *va, const void *vb)
{
    struct mtrr *a = *(struct mtrr **)va;
    struct mtrr *b = *(struct mtrr **)vb;

    if (a->base < b->base)
	return -1;
    else if (a->base > b->base)
	return +1;
    else if (a->size > b->size)
	return -1;
    else if (a->size < b->size)
	return +1;
    else
	return 0;
}

static struct mtrr **
find_children(struct mtrr *parent, struct mtrr **start)
{
    struct mtrr **d = start;
    struct mtrr *elder_sib = NULL;

    while (d < &sorted_mtrr[mtrrs_used]
    && inside(parent, d[0]->base))
    {
	struct mtrr *r = d[0];

	if (parent->eldest_child == NULL)
	    parent->eldest_child = r;
	r->parent = parent;
	if (elder_sib != NULL)
	    elder_sib->younger_sib = r;
	r->elder_sib = elder_sib;
	r->younger_sib = NULL;
	elder_sib = r;
	d = find_children(r, d+1);
    }
    return d;
}

static void
organize_mtrr(void)
{
    /* created ordered array of used MTRRs */
    {
	struct mtrr *r;
	struct mtrr **s = sorted_mtrr;

	for (r = mtrr; r != &mtrr[mtrrs_used]; r++)
	    *s++ = r;
    }
    qsort(sorted_mtrr, mtrrs_used, sizeof(sorted_mtrr[0]), reg_comp);
    assert(sorted_mtrr[0] == top_mtrr);	/* top MTRR should be first */

    {
	struct mtrr **r = find_children(mtrr, sorted_mtrr+1);

	assert(r == &sorted_mtrr[mtrrs_used]);
    }

    /* print sorted list, indented */
    if (debug) {
	printf("\n");
	print_mtrr_children(top_mtrr, 0);
    }
}

static struct mtrr *
new_mtrr(addr_t base, addr_t size)
{
    static int next_regno = 50 > NUM_VAR_MTRR? 50 : NUM_VAR_MTRR + 10;	/* must start larger than NUM_VAR_MTRR */
    struct mtrr *r = malloc(sizeof(struct mtrr));

    if (r == NULL) {
	fprintf(stderr, "%s: malloc failure (mtrr)\n", me);
	exit(1);
    }
    *r = blank_mtrr;
    r->regno = next_regno++;
    r->base = base;
    r->size = size;
    return r;
}

static void
remove_mtrr(struct mtrr *r)
{
    struct mtrr *old_end;
    struct mtrr *young_end;

    if (r->flags & mtrr_protect) {
	fprintf(stderr, "%s: attempting to remove MTRR %d which is forbidden\n"
	    , me, r->regno);
	exit(2);
    }

    if (r->eldest_child == NULL) {
	/* old_end ... young_end is an empty list! */
	old_end = r->younger_sib;
	young_end = r->elder_sib;
    } else {
	struct mtrr *c;

	old_end = r->eldest_child;
	/* reparent children and find youngest */
	for (c = old_end; c != NULL; c = c->younger_sib) {
	    c->parent = r->parent;
	    young_end = c;
	}
	old_end->elder_sib = r->elder_sib;
	young_end->younger_sib = r->younger_sib;
    }

    /* replace r in sib list with old_end ... young_end */
    if (r->elder_sib == NULL)
	r->parent->eldest_child = old_end;
    else
	r->elder_sib->younger_sib = old_end;
    if (r->younger_sib != NULL)
	r->younger_sib->elder_sib = young_end;

    if (debug) {
	printf("DELETE: ");
	print_mtrr(r, 0);
    }

    /* r is now disconnected.
     * We could free it, but not worth the bother.
     * It could be staticly allocated on malloced.
     */
    r->flags |= mtrr_deleted;
}

static void
split_mtrr(struct mtrr *r)
{
    const addr_t hsz = r->size/2;	/* half size */
    struct mtrr *u = new_mtrr(r->base+hsz, hsz);    /* second node */
    struct mtrr *c;

    if (r->flags & mtrr_protect) {
	fprintf(stderr, "%s: attempting to split MTRR %d which is forbidden\n"
	    , me, r->regno);
	exit(2);
    }
    if (debug) {
	printf("SPLIT: ");
	print_mtrr(r, 0);
    }
    /* find c, to become the first child of s */
    for (c = r->eldest_child; c != NULL && c->base < r->base+hsz; c = c->younger_sib) {
    }
    if (c != NULL) {
	if (r->eldest_child == c)
	    r->eldest_child = NULL;
	else
	    c->elder_sib->younger_sib = NULL;
	c->elder_sib = NULL;
    }
    r->size = hsz;
    r->flags |= mtrr_modified;
    u->type = r->type;
    u->parent = r->parent;
    u->eldest_child = c;
    for ( ; c != NULL; c = c->younger_sib)
	c-> parent = u;
    u->elder_sib = r;
    u->younger_sib = r->younger_sib;
    r->younger_sib = u;
    if (u->younger_sib != NULL)
	u->younger_sib->elder_sib = u;
    if (debug) {
	print_mtrr(r, 1);
	print_mtrr(u, 1);
    }
}

static void
uncover(struct mtrr *r)
{
    for (;;) {
	struct mtrr *p = r->parent;

	if (p == top_mtrr)
	    break;	/* nothing covering us */

	uncover(p);

	if (p->base == r->base && p->size == r->size)
	    break;	/* parent exactly covers us so it will be optimized out */
	split_mtrr(p);
    }
}

/* find smallest MTRR to be uncovered */
static void
selective_uncover(struct target_range *tr)
{
    struct mtrr *r = top_mtrr;
    addr_t lwb = tr->lwb;
    addr_t upb = tr->upb;
    struct mtrr *best = NULL;
    struct mtrr *c = r->eldest_child;

    for (;;) {
	if (c == NULL) {
	    /* none left: give up */
	    break;
	} else if (lwb < c->base) {
	    /* bad, and siblings won't work either */
	    break;
	} else if (c->base + c->size <= upb) {
	    /* bad: try sibling */
	    c = c->younger_sib;
	} else {
	    /* good; now try children */
	    best = c;
	    c = c->eldest_child;
	}
    }

    if (best == NULL) {
	fprintf(stderr, "%s: warning no target MTRR found for [" AF "," AF ")\n"
	    , me, lwb, upb);
    } else {
	best->flags |= mtrr_target;
	if (debug) {
	    printf("TARGET ");
	    print_mtrr(best, 0);
	}
	uncover(best);
	tr->mtrr = best;
    }
}

static void
general_uncover(struct mtrr *r)
{
    /* Note: the tree structure will change as we traverse it.
     * The only ill effect is that some nodes will be visited
     * more than once.
     */
    struct mtrr *c;

    for (c = r->eldest_child; c != NULL; c = c->younger_sib) {
	uncover(c);
	general_uncover(c);
    }
}

/* From AMD64 Architecture Programmer's Manual, Volume 2, 7.7.4:
 * If two or more variable-range MTRRs overlap, the following rules apply:
 * a. If the memory types are identical, then that memory type is used.
 * b. If at least one of the memory types is UC, the UC memory type is used.
 * c. If at least one of the memory types is WT, and the only other memory
 *    type is WB, then the WT memory type is used.
 * d. If the combination of memory types is not listed Steps A through C
 *    immediately above, then the memory type used is undefined.
 */

/* This returns true after first optimization.
 * Optimization should be started again in this case:
 * any traversal state may have been broken by the optimization.
 */
static int
optimize(struct mtrr *r)
{
    unsigned int rt = r->type->type;
    struct mtrr *p = r->parent;
    struct mtrr *del = NULL;	/* to be deleted */

    /* optimize children first */
    {
	struct mtrr *c;

	for (c = r->eldest_child; c != NULL; c = c->younger_sib)
	    if (optimize(c))
		return TRUE;
    }

    if (p == NULL) {
	/* cannot optimize away top MTRR itself */
    } else {
	/* consider if parent or r is redundant because of the other */
	unsigned int pt = p->type->type;

	if (rt == pt
	&& !(r->flags & mtrr_protect)
	&& (p != top_mtrr || r->eldest_child == NULL))
	{
	    /* Same type, so r, which might be smaller, is redundant.
	     * Note top_mtrr is actually implicit and its direct
	     * children can only be redundant if they don't have
	     * children..
	     */
	    del = r;
	    if (debug)
		printf("child redundant: same type as parent ");
	} else if (p != top_mtrr && r->base == p->base && r->size == p->size) {
	    /* p isn't top and both are the same size.
	     * The type rules dictate which is redundant
	     */
	    if (rt == pt) {
		/* normally handled by earlier rt == pt case unless
		 * r was protected.  Let's try p instead.
		 */
		del = p;
	    } else if (pt == MTRR_TYPE_UNCACHABLE) {
		del = r;
		if (debug)
		    printf("child redundant: parent same size and parent UC ");
	    } else if (rt == MTRR_TYPE_UNCACHABLE) {
		del = p;
		if (debug)
		    printf("parent redundant: parent same size and child UC ");
	    } else if (pt == MTRR_TYPE_WRTHROUGH && rt == MTRR_TYPE_WRBACK) {
		del = r;
		if (debug)
		    printf("child redundant: parent WT and child WB ");
	    } else if (rt == MTRR_TYPE_WRTHROUGH && pt == MTRR_TYPE_WRBACK) {
		del = p;
		if (debug)
		    printf("parent redundant: parent WB and child WT ");
	    } else {
		fprintf(stderr, "%s: overlapping MTRRs %d and %d have conflicting types (%s and %s)\n"
		    , me, p->regno, p->regno, p->type->name, r->type->name);
		exit(2);
	    }
	}
    }
    if (del != NULL) {
	if (del->flags & mtrr_protect) {
	    if (debug)
		printf("but MTRR %d is protected\n", del->regno);
	} else {
	    remove_mtrr(del);
	    return TRUE;
	}
    }

    if (!(r->flags & mtrr_protect)) {

	/* Check if children render r redundant.
	 * Assumption: types passed earlier tests.
	 */
	{
	    struct mtrr *c;
	    addr_t a = r->base;

	    for (c = r->eldest_child; c != NULL; c = c->younger_sib) {
		if (a != c->base)
		    break;  /* there is a gap */
		a += c->size;
	    }
	    if (a == r->base + r->size) {
		if (debug)
		    printf("parent completely covered by children ");
		remove_mtrr(r);
		return TRUE;
	    }
	}

	/* attempt to consolidate with younger siblings */
	{
	    struct mtrr *s;	
	    addr_t sz = r->size;

	    for (s = r->younger_sib
	    ; s != NULL && !(s->flags & mtrr_protect) && r->type == s->type && r->base + sz == s->base
	    ; s = s->younger_sib) {
		sz += s->size;
		if (sz > lowbit(r->base))
		    break;	/* can't work: base not aligned enough */

		/* Before we consider grabbing, we'd better optimize
		 * This recursion is going to happen more times than
		 * needed, but that should not amount to much actual cost.
		 */
		if (optimize(s))
		    return TRUE;

		if (lowbit(sz) == sz) {
		    /* sz is a power of two: consolidate! */
		    struct mtrr *roof = s->younger_sib;
		    struct mtrr *yc = r->eldest_child;

		    r->size = sz;	/* claim the new territory */
		    r->flags |= mtrr_modified;

		    /* find r's youngest child (if any) */
		    if (yc != NULL)
			while (yc->younger_sib != NULL)
			    yc = yc->younger_sib;
		    while (r->younger_sib != roof) {
			/* steal all the children of the next younger sib */
			struct mtrr *t = r->younger_sib;
			struct mtrr *c = t->eldest_child;

			if (debug)
			    printf("consolodate %d with younger sibs ", r->regno);
			if (yc == NULL)
			    r->eldest_child = c;
			else
			    yc->younger_sib = c;
			if (c != NULL) {
			    c->elder_sib = yc;
			    for (; c != NULL; c = c->younger_sib) {
				c->parent = r;
				yc = c;
			    }
			}
			t->eldest_child = NULL;
			remove_mtrr(t);
		    }

		    return TRUE;
		}
	    }
	}
    }

    return FALSE;
}

static int changed = 0;
static int unchanged = 0;
static int added = 0;

static void
count_mtrrs(struct mtrr *r)
{
    /* Note we don't count our own node, only children.
     * This avoids counting top_mtrr.
     */
    struct mtrr *c;

    for (c = r->eldest_child; c != NULL; c = c->younger_sib) {
	if (c->regno < NUM_VAR_MTRR) {
	    if (c->flags & mtrr_modified)
		changed++;
	    else
		unchanged++;
	} else {
	    added++;
	}
	count_mtrrs(c);
    }
}

/* Report about changes we propose.
 * returns: anything changed?
 */
static int
report_statistics(void)
{
    /* check to see that the number of MTRRs required is small enough */

    count_mtrrs(top_mtrr);
    if (unchanged == mtrrs_used - 1 && added == 0) {
	printf("No changes made.\n");
	return FALSE;
    } else {
	int n = changed + unchanged + added;
	int tmd = 0;	/* number of target MTRRs deleted */
	struct target_range *tr;

	for (tr = targets; tr != NULL; tr = tr->previous)
	    if (tr->mtrr != NULL && (tr->mtrr->flags & mtrr_deleted))
		tmd++;

	if (n > NUM_VAR_MTRR) {
	    fprintf(stderr, "%s: %d MTRRs needed but only %d in architecture.\n"
		, me, n, NUM_VAR_MTRR);
	    exit(2);
	} else if (n + tmd > NUM_VAR_MTRR) {
	    fprintf(stderr, "%s: warning: %d MTTRs used and %d target MTRRs were deleted (which may be needed); architecture only provides %d.\n"
		, me, n, tmd, NUM_VAR_MTRR);
	}
	if (debug)
	    printf("\nMTRR usage goes from %d to %d out of %d\n"
		, mtrrs_used - 1, n, NUM_VAR_MTRR);
	return TRUE;
    }
}

static char cmd_buf[1000];	/* lots of room for the command sequence */
static size_t cmd_len = 0;

static void
add_cmd(ssize_t n)
{

    if (n < 0) {
	fprintf(stderr, "%s: error formatting command: %s\n"
	    , me, strerror(errno));
	exit(2);
    }
    cmd_len += n;
    if (cmd_len >= sizeof(cmd_buf) - 1) {
	fprintf(stderr, "%s: command output too long (more than %d characters)\n"
	    , me, (int)sizeof(cmd_buf) - 2);
	exit(2);
    }
}

static void
install_mtrrs(struct mtrr *r)
{
    /* We don't do our own node.
     * This avoids handling top_mtrr.
     */
    struct mtrr *c;

    for (c = r->eldest_child; c != NULL; c = c->younger_sib) {
	if (c->regno > NUM_VAR_MTRR || (c->flags & mtrr_modified)) {
	    add_cmd(snprintf(cmd_buf + cmd_len, sizeof(cmd_buf) - cmd_len
		, "base=" AF " size=" AF " type=%s\n", c->base, c->size, c->type->name));
	}
	install_mtrrs(c);
    }
}

static void
emit_commands(void)
{
    /* emit any required delete commands */
    {
	struct mtrr *r;

	for (r = top_mtrr + 1; r < &mtrr[mtrrs_used]; r++) {
	    if (r->flags & (mtrr_deleted | mtrr_modified)) {
		add_cmd(snprintf(cmd_buf + cmd_len, sizeof(cmd_buf) - cmd_len
		    , "disable=%d\n", r->regno));
	    }
	}
    }

    /* emit any required creations */
    install_mtrrs(top_mtrr);
}

static void
usage(void)
{
    fprintf(stderr, "usage: %s [--help] [--debug] [--mtrrfile mtrrfile] [--execute] [lwb[-upb]]...\n", me);
}

static void
bad_usage(const char *ugh)
{
    fprintf(stderr, "%s: %s\n", me, ugh);
    usage();
    exit(1);
}

int
main(int argc, char **argv)
{
    char **ap = argv;
    int execute = FALSE;

    me = ap[0];

    while (*++ap != NULL) {
	if (ap[0][0] == '-') {
	    if (strcmp(ap[0], "--help") == 0) {
		usage();
		return 0;
	    } else if (strcmp(ap[0], "--debug") == 0) {
		debug = TRUE;
	    } else if (strcmp(ap[0], "--mtrrfile") == 0) {
		if (ap[1] == NULL)
		    bad_usage("missing filename for --mtrrfile");
		else
		    pmp = *++ap;
	    } else if (strcmp(ap[0], "--execute") == 0) {
		execute = TRUE;
	    } else {
		bad_usage("unknown flag");
	    }
	} else {
	    unsigned long long n, m;
	    char *ep;
	    struct target_range *tr = malloc(sizeof(struct target_range));

	    if (tr == NULL) {
		fprintf(stderr, "%s: malloc failure (target_range)\n", me);
		exit(1);
	    }

	    errno = 0;
	    m = n = strtoull(ap[0], &ep, 0);
	    if (errno == 0 && *ep == '-')
		m = strtoull(ep+1, &ep, 0);
	    if (errno != 0 || *ep != '\0')
		bad_usage("expected number for bound");
	    if ((n|m) != (addr_t)(n|m))
		bad_usage("bounds too large");
	    tr->lwb = n;
	    tr->upb = m;
	    tr->mtrr = NULL;
	    tr->previous = targets;
	    targets = tr;
	}
    }

    slurp();
    organize_mtrr();
    if (!execute) {
	printf("Initial MTRR configuration:\n");
	print_mtrr_children(top_mtrr, 0);
    }
    if (targets == NULL) {
	general_uncover(top_mtrr);
    } else {
	struct target_range *tr = targets;

	while (tr != NULL) {
	    selective_uncover(tr);
	    tr = tr->previous;
	}
    }
    do {
	if (debug) {
	    printf("\n");
	    print_mtrr_children(top_mtrr, 0);
	}
    } while (optimize(top_mtrr));

    if (report_statistics()) {
	emit_commands();
	if (!execute) {
	    printf("\nFinal MTRR configuration:\n");
	    print_mtrr_children(top_mtrr, 0);
	}
	if (debug || !execute) {
	    printf("\nCommands for /proc/mtrr to make these changes:\n");
	    fwrite(cmd_buf, sizeof(char), (size_t) cmd_len, stdout);
	}

	if (execute) {
	    int cmdf = creat(pmp, S_IRUSR | S_IWUSR);

	    if (cmdf < 0) {
		fprintf(stderr, "%s: cannot open %s for writing: %s\n", me, pmp, strerror(errno));
		exit(1);
	    } else {
		/* Writing in one fell swoop seems to fail so we write one line at a time.
		 * We use bare I/O instead of stdio to make the path as simple as possible.
		 */
		char *s;
		int out_line_no = 0;

		for (s = cmd_buf; *s != '\0'; ) {
		    char *e = strchr(s, '\n');
		    ssize_t n;
		    ssize_t len;

		    assert(e != NULL);
		    len = e - s + 1;
		    out_line_no++;
		    n = write(cmdf, s, len);

		    if (n < 0) {
			fprintf(stderr, "%s: cannot write line %d to %s: %s\n"
			    , me, out_line_no, pmp, strerror(errno));
			exit(1);
		    }
		    if (n < len) {
			fprintf(stderr, "%s: could only write %d of %d characters of line %d to %s"
			    , me, (int) n, (int) len, out_line_no, pmp);
			exit(1);
		    }
		    s = e + 1;
		}
		if (close(cmdf) < 0) {
		    fprintf(stderr, "%s: cannot close %s: %s\n", me, pmp, strerror(errno));
		    exit(1);
		}
	    }
	}
    }

    return 0;
}
